@@ -1,6 +1,8 @@
 // Copyright © SixtyFPS GmbH <info@slint-ui.com>
 // SPDX-License-Identifier: GPL-3.0-only OR LicenseRef-Slint-commercial
 
+// cSpell: ignore hframe qreal tabbar vframe
+
 use i_slint_core::input::FocusEventResult;
 
 use super::*;
@@ -18,6 +20,8 @@ pub struct NativeTabWidget {
     pub content_min_width: Property<f32>,
     pub tabbar_preferred_height: Property<f32>,
     pub tabbar_preferred_width: Property<f32>,
+    pub current_index: Property<i32>,
+    pub current_focused: Property<i32>,
 
     // outputs
     pub content_x: Property<f32>,
@@ -313,8 +317,9 @@ pub struct NativeTab {
     pub enabled: Property<bool>,
     pub pressed: Property<bool>,
     pub current: Property<i32>,
-    pub num_tabs: Property<i32>,
+    pub current_focused: Property<i32>,
     pub tab_index: Property<i32>,
+    pub num_tabs: Property<i32>,
     pub cached_rendering_data: CachedRenderingData,
 }
 
@@ -439,6 +444,7 @@ impl Item for NativeTab {
         .unwrap_or_default();
         let enabled: bool = this.enabled();
         let current: i32 = this.current();
+        let current_focused: i32 = this.current_focused();
         let tab_index: i32 = this.tab_index();
         let num_tabs: i32 = this.num_tabs();
 
@@ -453,6 +459,7 @@ impl Item for NativeTab {
             dpr as "float",
             tab_index as "int",
             current as "int",
+            current_focused as "int",
             num_tabs as "int",
             initial_state as "int"
         ] {
@@ -481,6 +488,9 @@ impl Item for NativeTab {
             }
             if (current == tab_index)
                 option.state |= QStyle::State_Selected;
+            if (current_focused == tab_index) {
+                option.state |= QStyle::State_HasFocus | QStyle::State_KeyboardFocusChange | QStyle::State_Item;
+            }
             option.features |= QStyleOptionTab::HasFrame;
             qApp->style()->drawControl(QStyle::CE_TabBarTab, &option, painter, widget);
         });
