@@ -5,15 +5,17 @@
 
 use cpp::*;
 use euclid::approxeq::ApproxEq;
+use i_slint_core::graphics::rendering_metrics_collector::{
+    RenderingMetrics, RenderingMetricsCollector,
+};
 use i_slint_core::graphics::{
-    Brush, Color, FPSCounter, FontRequest, Image, Point, Rect, RenderingCache, SharedImageBuffer,
-    Size,
+    Brush, Color, FontRequest, Image, Point, Rect, RenderingCache, SharedImageBuffer, Size,
 };
 use i_slint_core::input::{KeyEvent, KeyEventType, MouseEvent};
 use i_slint_core::item_rendering::{CachedRenderingData, ItemRenderer};
 use i_slint_core::items::{
-    self, FillRule, ImageRendering, InputType, ItemRef, MouseCursor, PointerEventButton,
-    TextOverflow, TextWrap,
+    self, FillRule, ImageRendering, InputType, ItemRc, ItemRef, Layer, MouseCursor, Opacity,
+    PointerEventButton, RenderingResult, TextOverflow, TextWrap,
 };
 use i_slint_core::layout::Orientation;
 use i_slint_core::window::{PlatformWindow, PopupWindow, PopupWindowLocation, WindowRc};
@@ -71,6 +73,7 @@ cpp! {{
 
     struct SlintWidget : QWidget {
         void *rust_window;
+        bool isMouseButtonDown = false;
 
         SlintWidget() {
             setMouseTracking(true);
@@ -95,6 +98,7 @@ cpp! {{
         }
 
         void mousePressEvent(QMouseEvent *event) override {
+            isMouseButtonDown = true;
             QPoint pos = event->pos();
             int button = event->button();
             rust!(Slint_mousePressEvent [rust_window: &QtWindow as "void*", pos: qttypes::QPoint as "QPoint", button: u32 as "int" ] {
@@ -104,6 +108,21 @@ cpp! {{
             });
         }
         void mouseReleaseEvent(QMouseEvent *event) override {
+            // HACK: Qt on windows is a bit special when clicking on the window
+            //       close button and when the resulting close event is ignored.
+            //       In that case a release event that was not preceeded by
+            //       a press event is sent on Windows.
+            //       This confuses Slint, so eat this event.
+            //
+            //       One example is a popup is shown in the close event that
+            //       then ignores the the close request to ask the user what to
+            //       do. The stray release event will then close the popup
+            //       straight away
+            if (!isMouseButtonDown) {
+                return;
+            }
+            isMouseButtonDown = false;
+
             QPoint pos = event->pos();
             int button = event->button();
             rust!(Slint_mouseReleaseEvent [rust_window: &QtWindow as "void*", pos: qttypes::QPoint as "QPoint", button: u32 as "int" ] {
@@ -181,6 +200,20 @@ cpp! {{
             QWidget::changeEvent(event);
         }
 
+        void closeEvent(QCloseEvent *event) override {
+            bool accepted = rust!(Slint_requestClose [rust_window: &QtWindow as "void*"] -> bool as "bool" {
+                if let Some(window) = rust_window.self_weak.upgrade() {
+                    return window.request_close();
+                }
+                true
+            });
+            if (accepted) {
+                event->accept();
+            } else {
+                event->ignore();
+            }
+        }
+
         QSize sizeHint() const override {
             auto preferred_size = rust!(Slint_sizeHint [rust_window: &QtWindow as "void*"] -> qttypes::QSize as "QSize" {
                 let component_rc = rust_window.self_weak.upgrade().unwrap().component();
@@ -206,6 +239,7 @@ cpp! {{
     static int do_text_layout(QTextLayout &layout, int flags, const QRectF &rect, int line_for_y_pos = -1) {
         QTextOption options;
         options.setWrapMode((flags & Qt::TextWordWrap) ? QTextOption::WordWrap : QTextOption::NoWrap);
+        options.setFlags(QTextOption::IncludeTrailingSpaces);
         layout.setTextOption(options);
         layout.setCacheEnabled(true);
         QFontMetrics fm(layout.font());
@@ -372,6 +406,7 @@ struct QtItemRenderer<'a> {
     cache: QtRenderingCache,
     default_font_properties: FontRequest,
     window: WindowRc,
+    metrics: RenderingMetrics,
 }
 
 impl ItemRenderer for QtItemRenderer<'_> {
@@ -762,6 +797,25 @@ impl ItemRenderer for QtItemRenderer<'_> {
         }}
     }
 
+    fn visit_opacity(&mut self, opacity_item: Pin<&Opacity>, self_rc: &ItemRc) -> RenderingResult {
+        let opacity = opacity_item.opacity();
+        if Opacity::need_layer(self_rc, opacity) {
+            self.render_and_blend_layer(&opacity_item.cached_rendering_data, opacity, self_rc)
+        } else {
+            self.apply_opacity(opacity);
+            opacity_item.cached_rendering_data.release(&mut self.cache.borrow_mut());
+            RenderingResult::ContinueRenderingChildren
+        }
+    }
+
+    fn visit_layer(&mut self, layer_item: Pin<&Layer>, self_rc: &ItemRc) -> RenderingResult {
+        if layer_item.cache_rendering_hint() {
+            self.render_and_blend_layer(&layer_item.cached_rendering_data, 1.0, self_rc)
+        } else {
+            RenderingResult::ContinueRenderingChildren
+        }
+    }
+
     fn combine_clip(&mut self, rect: Rect, radius: f32, mut border_width: f32) {
         let mut clip_rect = qttypes::QRectF {
             x: rect.min_x() as _,
@@ -861,6 +915,10 @@ impl ItemRenderer for QtItemRenderer<'_> {
             painter->setOpacity(painter->opacity() * opacity);
         }}
     }
+
+    fn metrics(&self) -> RenderingMetrics {
+        self.metrics.clone()
+    }
 }
 
 pub(crate) fn load_image_from_resource(
@@ -1100,6 +1158,95 @@ impl QtItemRenderer<'_> {
             }
         }}
     }
+
+    fn render_layer(
+        &mut self,
+        item_cache: &CachedRenderingData,
+        item_rc: &ItemRc,
+        layer_size_fn: &dyn Fn() -> qttypes::QSize,
+    ) -> Option<qttypes::QPixmap> {
+        let cache_entry = item_cache.get_or_update(&self.cache.clone(), || {
+            let layer_size: qttypes::QSize = layer_size_fn();
+            let mut layer_image = qttypes::QImage::new(layer_size, qttypes::ImageFormat::ARGB32_Premultiplied);
+            layer_image.fill(qttypes::QColor::from_rgba_f(0., 0., 0., 0.));
+
+            *self.metrics.layers_created.as_mut().unwrap() += 1;
+
+            let mut layer_painter = {
+                let img_ref: &mut qttypes::QImage = &mut layer_image;
+                cpp!(unsafe [img_ref as "QImage*"] -> QPainter as "QPainter" { return QPainter(img_ref); })
+            };
+
+            std::mem::swap(self.painter, &mut layer_painter);
+
+            {
+                let painter: &mut QPainter = &mut *self.painter;
+                cpp! { unsafe [
+                        painter as "QPainter*",
+                        layer_size as "QSize"
+                    ] {
+                    painter->setClipRect(0, 0, layer_size.width(), layer_size.height());
+                }}
+            }
+
+            i_slint_core::item_rendering::render_item_children(
+                self,
+                &item_rc.component(),
+                item_rc.index() as isize,
+            );
+
+            std::mem::swap(self.painter, &mut layer_painter);
+            drop(layer_painter);
+
+            let img_ref = &mut layer_image;
+            QtRenderingCacheItem::Pixmap(cpp!(unsafe [img_ref as "QImage*"] -> qttypes::QPixmap as "QPixmap" { return QPixmap::fromImage(*img_ref); }))
+        });
+        match &cache_entry {
+            QtRenderingCacheItem::Pixmap(pixmap) => Some(pixmap.clone()),
+            _ => None,
+        }
+    }
+
+    fn render_and_blend_layer(
+        &mut self,
+        item_cache: &CachedRenderingData,
+        alpha_tint: f32,
+        self_rc: &ItemRc,
+    ) -> RenderingResult {
+        let current_clip = self.get_current_clip();
+        if let Some(mut layer_image) = self.render_layer(&item_cache, self_rc, &|| {
+            // We don't need to include the size of the opacity item itself, since it has no content.
+            let children_rect = i_slint_core::properties::evaluate_no_tracking(|| {
+                let self_ref = self_rc.borrow();
+                self_ref.as_ref().geometry().union(
+                    &i_slint_core::item_rendering::item_children_bounding_rect(
+                        &self_rc.component(),
+                        self_rc.index() as isize,
+                        &current_clip,
+                    ),
+                )
+            });
+            qttypes::QSize {
+                width: children_rect.size.width as _,
+                height: children_rect.size.height as _,
+            }
+        }) {
+            self.save_state();
+            self.apply_opacity(alpha_tint);
+            {
+                let painter: &mut QPainter = &mut *self.painter;
+                let layer_image_ref: &mut qttypes::QPixmap = &mut layer_image;
+                cpp! { unsafe [
+                        painter as "QPainter*",
+                        layer_image_ref as "QPixmap*"
+                    ] {
+                    painter->drawPixmap(0, 0, *layer_image_ref);
+                }}
+            }
+            self.restore_state();
+        }
+        RenderingResult::ContinueRenderingWithoutChildren
+    }
 }
 
 cpp_class!(unsafe struct QWidgetPtr as "std::unique_ptr<QWidget>");
@@ -1108,7 +1255,7 @@ pub struct QtWindow {
     widget_ptr: QWidgetPtr,
     pub(crate) self_weak: Weak<i_slint_core::window::Window>,
 
-    fps_counter: Option<Rc<FPSCounter>>,
+    rendering_metrics_collector: Option<Rc<RenderingMetricsCollector>>,
 
     cache: QtRenderingCache,
 }
@@ -1122,7 +1269,7 @@ impl QtWindow {
         let rc = Rc::new(QtWindow {
             widget_ptr,
             self_weak: window_weak.clone(),
-            fps_counter: FPSCounter::new(),
+            rendering_metrics_collector: RenderingMetricsCollector::new(window_weak.clone()),
             cache: Default::default(),
         });
         let self_weak = Rc::downgrade(&rc);
@@ -1150,6 +1297,7 @@ impl QtWindow {
                 cache,
                 default_font_properties: self.default_font_properties(),
                 window: runtime_window,
+                metrics: RenderingMetrics { layers_created: Some(0) },
             };
 
             for (component, origin) in components {
@@ -1160,8 +1308,8 @@ impl QtWindow {
                 );
             }
 
-            if let Some(fps_counter) = &self.fps_counter {
-                fps_counter.measure_frame_rendered(&mut renderer);
+            if let Some(collector) = &self.rendering_metrics_collector {
+                collector.measure_frame_rendered(&mut renderer);
             }
 
             i_slint_core::animations::CURRENT_ANIMATION_DRIVER.with(|driver| {
@@ -1238,11 +1386,11 @@ impl PlatformWindow for QtWindow {
         cpp! {unsafe [widget_ptr as "QWidget*"] {
             widget_ptr->show();
         }};
-        if let Some(fps_counter) = &self.fps_counter {
+        if let Some(collector) = &self.rendering_metrics_collector {
             let qt_platform_name = cpp! {unsafe [] -> qttypes::QString as "QString" {
                 return QGuiApplication::platformName();
             }};
-            fps_counter.start(&format!("Qt backend (platform {})", qt_platform_name));
+            collector.start(&format!("Qt backend (platform {})", qt_platform_name));
         }
     }
 
@@ -1250,6 +1398,9 @@ impl PlatformWindow for QtWindow {
         let widget_ptr = self.widget_ptr();
         cpp! {unsafe [widget_ptr as "QWidget*"] {
             widget_ptr->hide();
+            // Since we don't call close(), this will force Qt to recompute wether there are any
+            // visible windows, and ends the application if needed
+            QEventLoopLocker();
         }};
     }
 
@@ -1278,8 +1429,7 @@ impl PlatformWindow for QtWindow {
         };
         if size.width == 0 || size.height == 0 {
             let existing_size = cpp!(unsafe [widget_ptr as "QWidget*"] -> qttypes::QSize as "QSize" {
-                auto sizeHint = widget_ptr->sizeHint();
-                return sizeHint.isValid() ? sizeHint : widget_ptr->size();
+                return widget_ptr->size();
             });
             if size.width == 0 {
                 window_item.width.set(existing_size.width as _);
@@ -1380,7 +1530,7 @@ impl PlatformWindow for QtWindow {
         let widget_ptr = self.widget_ptr();
         cpp! {unsafe [widget_ptr as "QWidget*", popup_ptr as "QWidget*", pos as "QPoint", size as "QSize"] {
             popup_ptr->setParent(widget_ptr, Qt::Popup);
-            popup_ptr->setGeometry(QRect(pos + widget_ptr->geometry().topLeft(), size));
+            popup_ptr->setGeometry(QRect(pos + widget_ptr->mapToGlobal(QPoint(0,0)), size));
             popup_ptr->show();
         }};
     }
@@ -1399,6 +1549,7 @@ impl PlatformWindow for QtWindow {
             MouseCursor::text => key_generated::Qt_CursorShape_IBeamCursor,
             MouseCursor::alias => key_generated::Qt_CursorShape_DragLinkCursor,
             MouseCursor::copy => key_generated::Qt_CursorShape_DragCopyCursor,
+            MouseCursor::r#move => key_generated::Qt_CursorShape_DragMoveCursor,
             MouseCursor::no_drop => key_generated::Qt_CursorShape_ForbiddenCursor,
             MouseCursor::not_allowed => key_generated::Qt_CursorShape_ForbiddenCursor,
             MouseCursor::grab => key_generated::Qt_CursorShape_OpenHandCursor,
@@ -1437,6 +1588,9 @@ impl PlatformWindow for QtWindow {
         text_input: Pin<&i_slint_core::items::TextInput>,
         pos: Point,
     ) -> usize {
+        if pos.y < 0. {
+            return 0;
+        }
         let rect: qttypes::QRectF = get_geometry!(items::TextInput, text_input);
         let pos = qttypes::QPointF { x: pos.x as _, y: pos.y as _ };
         let font: QFont =
@@ -1469,12 +1623,15 @@ impl PlatformWindow for QtWindow {
             QTextLayout layout(copy, font);
             auto line = do_text_layout(layout, flags, rect, pos.y());
             if (line < 0 || layout.lineCount() <= line)
-                return 0;
+                return string.toUtf8().size();
             QTextLine textLine = layout.lineAt(line);
             int cur;
             if (pos.x() > textLine.naturalTextWidth()) {
                 cur = textLine.textStart() + textLine.textLength();
-                if (cur > 0 && string[cur - 1] == '\n')
+                // cur is one past the last character of the line (eg, the \n or space).
+                // Go one back to get back on this line.
+                // Unless we were at the end of the text, in which case there was no \n
+                if (cur > textLine.textStart() && (cur < string.size() || string[cur-1] == '\n'))
                     cur--;
             } else {
                 cur = textLine.xToCursor(pos.x());
@@ -1486,11 +1643,11 @@ impl PlatformWindow for QtWindow {
         }}
     }
 
-    fn text_input_position_for_byte_offset(
+    fn text_input_cursor_rect_for_byte_offset(
         &self,
         text_input: Pin<&i_slint_core::items::TextInput>,
         byte_offset: usize,
-    ) -> Point {
+    ) -> Rect {
         let rect: qttypes::QRectF = get_geometry!(items::TextInput, text_input);
         let font: QFont =
             get_font(text_input.unresolved_font_request().merge(&self.default_font_properties()));
@@ -1523,7 +1680,13 @@ impl PlatformWindow for QtWindow {
                 return QPointF();
             return QPointF(textLine.x() + textLine.cursorToX(offset), textLine.y());
         }};
-        Point::new(r.x as _, r.y as _)
+
+        let font_size = cpp! { unsafe [font as "QFont"]
+                -> i32 as "int" {
+            return QFontInfo(font).pixelSize();
+        }};
+
+        Rect::new(Point::new(r.x as _, r.y as _), Size::new(1.0, font_size as f32))
     }
 
     fn as_any(&self) -> &dyn std::any::Any {
